/************************************************************************

PlayOGGData.cpp

Copyright (C) 2008 Micket

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

************************************************************************/
#pragma warning(disable:4786)

#include "CDDAData.hpp"
#include "Utils.hpp"

using namespace std;

extern Preferences prefs;

// this callback repeats one track over and over
void OggCallback(void * userData, Uint8 * outputBuffer, int len)
{
    // Cast data passed through stream to our structure type.
    PlayOGGData* data = (PlayOGGData*)userData;
    if(data->oggLoaded)
    {
        long read = 0;
        for (int left = len; left > 0; left = len - read)
        {
            long read2 = ov_read(&(data->vf),(char*)outputBuffer+read, 
                left, 0, 2, 1, &(data->current_section));
            if (read2 > 0)
                read += read2;
            else if (read2 < 0)
                fprintf(stderr,"Error %d\n",read2);
            else
            {
                if (data->repeat == 0) // Play one
                {
                    ov_clear(&(data->vf));
                    data->oggLoaded = false;
                    memset((char*)outputBuffer+read,left,0);
                    data->CDDAPos = data->trackList[data->track].trackEnd;
                    data->endOfTrack = true;
                    return;
                }
                else if (data->repeat == 1) // Loop one
                {
                    data->CDDAPos = data->trackList[data->track].trackStart;
                    ov_time_seek(&(data->vf),0);
                    data->current_section = 0;
                }
                else
                {
                    int tries = 0;
                    do 
                    {
                        tries++;
                        data->track++;
                        if (data->track >= data->trackList.size())
                            data->track = 2;
                        data->loadTrack();
                        
                        // One way through all tracks, then give up
                        if (tries >= data->trackList.size()-2)
                        {
                            data->playing = false;
                            data->endOfTrack = true;
                            memset((char*)outputBuffer+read,left,0);
                            return;
                        }
                    } while (!data->oggLoaded);
                }
            }
        }
        short *out = (short*)outputBuffer;
        for (int i = 0; i < len>>1; i++)
            out[i] = (short)(out[i] * data->volume);
    }
    else
        memset((char*)outputBuffer,len,0);
    double seconds = ((double)(len>>1))/(data->channels*data->freq);
    data->CDDAPos += CDTime((int)(seconds/60),((int)seconds)%60,
                            (seconds-(int)seconds)*75);
    return;
}

bool PlayOGGData::loadTrack()
{
    if (oggLoaded) ov_clear(&vf);
    int realTrack = track-1; // I thought it was nicer to number them from 1.
    gchar* songName = g_strdup_printf("%s.%02d.ogg", fileName.c_str(), realTrack);
    bool ok = ov_fopen(songName,&vf)==0;
    //printf("PlayOGGData::loadTrack(), track = %d)\n",realTrack);
    if (!ok)
        fprintf(stderr,"PlayOGGData::loadTrack(%d) - %s not found\n",realTrack,songName);
    else
        fprintf(stdout,"PlayOGGData::loadTrack(%d) - %s found\n",realTrack,songName);
    g_free(songName);
    CDDAPos = trackList[track].trackStart;
    vorbis_info *vi = ov_info(&vf,-1);
    if (ok && audioOpen)
    {
        if (vi->rate != freq || vi->channels != channels)
        fprintf(stderr,"PlayOGGData::loadTrack(%d) - Error: "
            "Songs should have same samplerate and channels, file: %s\n",
            realTrack, songName);
    }
    else
    {
        freq       = vi->rate;
        channels   = vi->channels;
    }
    return oggLoaded = ok;
}

bool PlayOGGData::openAudio()
{
    SDL_AudioSpec *desired = (SDL_AudioSpec*)malloc(sizeof(SDL_AudioSpec));
    desired->freq     = freq;
    desired->channels = channels;
    desired->format   = AUDIO_S16SYS;
    desired->samples  = 4096;
    desired->callback = OggCallback;
    desired->userdata = (void*)this;
    bool ok = SDL_OpenAudio(desired,NULL) == 0;
    free(desired);
    if (!ok)
        fprintf(stderr, "Couldn't open audio: %s\n", SDL_GetError());
    return ok;
}

PlayOGGData::PlayOGGData(const std::vector<TrackInfo> ti)
    : trackList(ti), playing(false), audioOpen(false), 
        oggLoaded(false), endOfTrack(false)
{
    //printf("PlayOGGData::PlayOGGData\n");
    setVolume(prefs.prefsMap[volumeString]);
    setRepeat(prefs.prefsMap[repeatString]);
}

// initialize the CDDA file data and initalize the audio stream
void PlayOGGData::openFile(const std::string& file) 
{
    //printf("PlayOGGData::openFile(%s)\n",file.c_str());
    fileName = stripFileExt(file);
    if(SDL_Init(SDL_INIT_AUDIO)==-1)
    {
        Exception e(string("SDL_Init: ") + string(SDL_GetError()));
        THROW(e);
    }
}

// start playing the data
int PlayOGGData::play(const CDTime& startTime)
{
    //printf("PlayOGGData::play\n");
    current_section = 0;
    endOfTrack = false;
    if (playing)
    {
        if (startTime == initialTime)
            return 0;
        else
            this->stop();
    }

    initialTime = startTime;

    // figure out which track to play to set the end time
    track = 2;
    while ( track < trackList.size()-1 && startTime > trackList[track+1].trackStart )
        track++;

    CDTime adjustedStartTime = startTime;
    // These 5 seconds have *nothing* to do with pregap
    CDTime diff(0,5,0);

    if (adjustedStartTime < (trackList[track].trackStart + diff))
        adjustedStartTime = trackList[track].trackStart;
    // If very close to the next track, adjust to that instead
    // I'm not sure i *want*  to do this though? Maybe the game wants to have a seconds
    // gap (Wipeout consistenly requests times that are around a second 
    // less then the track start )
    else if (track+1 < trackList.size() && 
        adjustedStartTime > trackList[track+1].trackStart - diff)
    {
        track++;
        adjustedStartTime = trackList[track].trackStart;
    }

    if (!loadTrack()) return -1;
    if (!audioOpen && !(audioOpen  = openAudio())) 
    {
        if (oggLoaded) ov_clear(&vf);
        return -1;
    }

    playing    = true;
    SDL_PauseAudio(0);
    return 0;
}

// close the stream - nice and simple

int PlayOGGData::stop()
{
    //printf("PlayOGGData::stop\n");
    if (audioOpen)
    {
        SDL_CloseAudio();
        audioOpen = false;
    }
    if (oggLoaded) ov_clear(&vf);
    playing = false;
    return 0;
}

